/** FuzzyTreeViewer.java
 *  Copyright (c) 2001 Donna Malayeri
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.JRadioButton;
import java.awt.*;
import javax.swing.border.*;
import java.awt.event.*;
import java.io.File;

import edu.umd.cs.fuzzyTree.*;

/** Application that uses the FuzzyTree and MovingSlider.  Allows user to open a new
 *  XML file, set FuzzyTree parameters, and change the number of nodes that are displayed.
 *
 *  @author     Donna Malayeri
 *  @version    24-Jul-01
 */

public class FuzzyTreeViewer extends JFrame
{
	FuzzyTree theTree;
	MovingSlider slider;
	File currentDirectory;

	private static final String TITLE = "Fuzzy Tree XML Viewer";

	private static final int MIN_SLIDER_VAL = (Integer.getInteger("minSliderVal", 0)).intValue();
	private static final int DISPLAY_RANGE = (Integer.getInteger("sliderDisplayRange", 50)).intValue();
	private static final int LEFT_MARGIN = (Integer.getInteger("sliderLeftMargin", 10)).intValue();
	private static final int RIGHT_MARGIN = (Integer.getInteger("sliderRightMargin", 10)).intValue();

	private static final int INITIAL_DISPLAY_NODES = (Integer.getInteger("initialDisplayedNodes", 15)).intValue();

	GridBagLayout gridBagLayout1 = new GridBagLayout();
	JPanel panel1 = new JPanel();
	TitledBorder titledBorder1;
	JTextField numNodesTextField = new JTextField();
	GridBagLayout gridBagLayout2 = new GridBagLayout();
	JLabel numNodesLabel = new JLabel();
	JButton numNodesButton = new JButton();
	JPanel panel2 = new JPanel();
	Border border1;
	JLabel displayedNodesLabel = new JLabel();
	GridBagLayout gridBagLayout3 = new GridBagLayout();
	JButton changeParametersButton = new JButton();
	JButton helpButton = new JButton();
	JScrollPane scrollPane;
	JLabel totalNodesLabel = new JLabel();
	JButton openButton = new JButton();

	public void initFrame() throws Exception
	{
		setTitle(TITLE);

		// add window listener
		addWindowListener(new WindowAdapter()
		{
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});

		changeParametersButton.addActionListener(new ChangeParametersListener());
		NumNodesListener listener = new NumNodesListener();
		numNodesTextField.addActionListener(listener);
		numNodesButton.addActionListener(listener);
		slider.addChangeListener(new SliderListener());

		jbInit();
		pack();
	}

	public FuzzyTreeViewer() throws Exception
	{
		slider = new MovingSlider();

		initFrame();
		setControlsEnabled(false);
	}

	public FuzzyTreeViewer(String filename) throws Exception
	{
		theTree = new FuzzyTree(filename);
		theTree.setNumDisplayedNodes(Math.min(INITIAL_DISPLAY_NODES, theTree.getTotalNodes()));

		slider = new MovingSlider(MIN_SLIDER_VAL, theTree.getTotalNodes(), theTree.getNumDisplayedNodes(),
									DISPLAY_RANGE, LEFT_MARGIN, RIGHT_MARGIN);
		displayElementInformation();

		initFrame();

		showFilenameInTitle(filename);
	}

	/** Listener for click on "Change Parameters" button */
	class ChangeParametersListener implements ActionListener
	{
		public void actionPerformed(ActionEvent e)
		{
			ParametersDialog dialog =
				new ParametersDialog(FuzzyTreeViewer.this, theTree.getParameters());
			dialog.show();

			if (!dialog.wasCanceled)
				theTree.setParameters(dialog.getParameters());
		}
	}

	/** Listener for changing the number of nodes using the text field */
	class NumNodesListener implements ActionListener
	{
		class RangeException extends Exception
		{
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				int newValue = Integer.parseInt(numNodesTextField.getText());
				if (newValue < slider.getRealMinimum() || newValue > slider.getRealMaximum())
					throw new RangeException();
				slider.adjustValue(newValue);
				numNodesTextField.setText("");
			}

			catch (Exception exception)
			{
				String errorMessage =
					"Enter an integer between " + slider.getRealMinimum() + " and " + slider.getRealMaximum();

				JOptionPane.showMessageDialog(FuzzyTreeViewer.this, errorMessage, "Error", JOptionPane.ERROR_MESSAGE);
				numNodesTextField.selectAll();
				numNodesTextField.requestFocus();
			}
		}
	}

	/** Listener for a change in slider value */
	class SliderListener implements ChangeListener
	{
		public void stateChanged(ChangeEvent e)
		{
			JSlider source = (JSlider)e.getSource();
			if (!source.getValueIsAdjusting())
			{
				int numNodes = (int)source.getValue();

				if (numNodes > 0)
				{
					theTree.setNumDisplayedNodes(numNodes);
					displayedNodesLabel.setText("Displayed: " + numNodes);
				}

				else
					source.setValue(1);
			}

		}
	}

	// for NumNodesChangedListener interface
	public void numNodesChanged(int newValue)
	{
		slider.setValue(newValue);
	}

	public static void main(String[] args)
	{
		try
		{
			JFrame mainFrame = null;

			if (args.length == 1)
			{
				if (args[0].equals("-help"))
				{
					System.out.println( "Usage: FuzzyTreeViewer <xml document>\n\n" +
										"Enviroment variable options (set using -DoptionName=value):\n" +
										"initialDisplayedNodes      initial number of nodes displayed in the fuzzy tree\n" +
										"minSliderVal               minimum value for slider\n" +
										"sliderDisplayRange         range of values displayed by the slider\n" +
										"sliderLeftMargin           left margin for slider\n" +
										"sliderRightMargin          right margin for slider\n");
					System.exit(1);
				}

				else
					mainFrame = new FuzzyTreeViewer(args[0]);
			}

			else
				mainFrame = new FuzzyTreeViewer();

			mainFrame.setVisible(true);
		}

		catch (Exception e)
		{
			System.err.println(e.getMessage());
		}
	}

	// generated by JBuilder, but can also be edited by hand
	private void jbInit() throws Exception
	{
		scrollPane = new JScrollPane(theTree);

		titledBorder1 = new TitledBorder("");
		border1 = BorderFactory.createBevelBorder(BevelBorder.LOWERED,Color.white,Color.white,new Color(128, 134, 151),new Color(89, 93, 105));
		this.getContentPane().setLayout(gridBagLayout1);
		panel1.setBorder(titledBorder1);
		panel1.setLayout(gridBagLayout2);
		titledBorder1.setTitle("Displayed Elements");
		numNodesLabel.setText("Number to display: ");
		numNodesTextField.setToolTipText("New number of elements to display");
		numNodesTextField.setColumns(5);
		numNodesButton.setToolTipText("Set new number of elements to display");
		numNodesButton.setText("Set Value");
		panel2.setLayout(gridBagLayout3);
		changeParametersButton.setToolTipText("Change the parameters used by the fuzzy tree");
		changeParametersButton.setText("Change Parameters");
		helpButton.setText("Help");
		helpButton.addActionListener(new java.awt.event.ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				helpButton_actionPerformed(e);
			}
		});
		slider.setMinorTickSpacing(1);
		slider.setMajorTickSpacing(5);
		slider.setPaintLabels(true);
		slider.setPaintTicks(true);
		slider.setSnapToTicks(true);
		totalNodesLabel.setToolTipText("");
		scrollPane.setPreferredSize(new Dimension(550, 400));
		openButton.setToolTipText("Open a new XML document");
		openButton.setText("Open New File");
		openButton.addActionListener(new java.awt.event.ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				openButton_actionPerformed(e);
			}
		});
		this.getContentPane().add(panel1, new GridBagConstraints(0, 2, 2, 1, 1.0, 0.0
			,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL, new Insets(0, 6, 6, 6), 0, 0));
		panel1.add(slider, new GridBagConstraints(1, 0, GridBagConstraints.REMAINDER, 1, 1.0, 0.0
			,GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(5, 6, 5, 6), 0, 0));
		panel1.add(numNodesLabel, new GridBagConstraints(1, 1, 1, 1, 0.0, 0.0
			,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE, new Insets(6, 6, 0, 0), 0, 5));
		panel1.add(numNodesTextField, new GridBagConstraints(2, 1, 1, 2, 0.0, 0.0
			,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE, new Insets(6, 6, 6, 0), 0, 6));
		panel1.add(numNodesButton, new GridBagConstraints(3, 1, 1, 2, 0.0, 0.0
			,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE, new Insets(6, 6, 0, 0), 0, 0));
		panel1.add(displayedNodesLabel, new GridBagConstraints(4, 1, 1, 2, 1.0, 0.0
			,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE, new Insets(6, 0, 0, 6), 0, 5));
		panel1.add(totalNodesLabel, new GridBagConstraints(5, 1, 1, 1, 0.0, 0.0
			,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE, new Insets(6, 5, 0, 6), 0, 5));
		this.getContentPane().add(panel2, new GridBagConstraints(0, 0, 2, 1, 1.0, 0.0
			,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL, new Insets(6, 6, 0, 6), 0, 0));
		panel2.add(changeParametersButton, new GridBagConstraints(2, 0, 1, 1, 0.0, 0.0
			,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE, new Insets(0, 0, 0, 6), 0, 0));
		panel2.add(helpButton, new GridBagConstraints(3, 0, 1, 1, 1.0, 0.0
			,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE, new Insets(0, 0, 0, 0), 0, 0));
		panel2.add(openButton, new GridBagConstraints(1, 0, 1, 1, 0.0, 0.0
			,GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(0, 0, 0, 6), 0, 0));
		this.getContentPane().add(scrollPane, new GridBagConstraints(0, 1, 1, 1, 1.0, 1.0
			,GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(6, 6, 6, 6), 0, 0));

	}

	void helpButton_actionPerformed(ActionEvent e)
	{
		String helpStr = "Move slider or type new value to change number of elements to display\n" +
						 "Click on an element or the text to partially expand/collapse\n\n" +
						 "Control-Click to collapse partially expanded nodes (default is expand)\n" +
						 "Shift-Click to fully expand/collapse\n\n" +
						 "Elements with a different icon are elements that are displayed under an ancestor\n" +
						 "   that is not the element's parent";

		JOptionPane.showMessageDialog(FuzzyTreeViewer.this, helpStr, "Help", JOptionPane.INFORMATION_MESSAGE);

	}

	void openButton_actionPerformed(ActionEvent e)
	{
		JFileChooser fileChooser = new JFileChooser(currentDirectory);
		ExampleFileFilter filter = new ExampleFileFilter();
		filter.addExtension("xml");
		filter.setDescription("XML Documents");
		fileChooser.setFileFilter(filter);

		int returnVal = fileChooser.showOpenDialog(this);

		if (returnVal == fileChooser.APPROVE_OPTION)
		{
			// file://localhost/ needed so it works on Windows machines -- don't know why
			String filename = "file://localhost/" + fileChooser.getSelectedFile().getAbsolutePath();
			currentDirectory = fileChooser.getCurrentDirectory();

			try
			{
				if (theTree == null)
				{
					theTree = new FuzzyTree(filename);
					scrollPane.setViewportView(theTree);
					setControlsEnabled(true);
				}
				else
					theTree.setXMLFile(filename);

				showFilenameInTitle(filename);
				theTree.setNumDisplayedNodes(Math.min(INITIAL_DISPLAY_NODES,
					theTree.getTotalNodes()));

				slider.setValueIsAdjusting(true);
				slider.setDisplayRange(Math.min(DISPLAY_RANGE, theTree.getTotalNodes()));
				slider.setRealMaximum(theTree.getTotalNodes());
				slider.adjustValue(theTree.getNumDisplayedNodes());
				slider.setValueIsAdjusting(false);

				displayElementInformation();
			}

			catch (DataGraph.DataGraphException ex)
			{
				JOptionPane.showMessageDialog(this, ex.getMessage(), "Error loading file",
					JOptionPane.ERROR_MESSAGE);
			}

		}
	}

	/** Enables or disables the text field, slider, and change parameters button,
	 *  based on <code>value</code>.
	 */
	private void setControlsEnabled(boolean value)
	{
		numNodesTextField.setEnabled(value);
		slider.setEnabled(value);
		changeParametersButton.setEnabled(value);
		numNodesButton.setEnabled(value);
	}

	/** Displays the number of elements displayed, and the total number of elements in the tree */
	private void displayElementInformation()
	{
		totalNodesLabel.setText("Total: " + theTree.getTotalNodes());
		displayedNodesLabel.setText("Displayed: " + theTree.getNumDisplayedNodes());
	}

	/** Appends the name of the open XML document to the title bar */
	private void showFilenameInTitle(String filename)
	{
		File file = new File(filename);
		setTitle(TITLE + " - " + file.getName());
	}

}
