/** DataGraph.java
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package edu.umd.cs.fuzzyTree;

import java.lang.*;
import java.util.*;
import org.w3c.dom.*;
import java.io.*;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.DocumentBuilder;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/** Holds the relationship information of tree/graph.  The Xerces parser is used
 *  to generate a graph from an XML document.
 *
 *  @author     Thomas Baby
 *  @author     Donna Malayeri
 *  @version    24-Jul-01
 */

public class DataGraph
{
	public class DataGraphException extends Exception
	{
		public DataGraphException(Exception e)
		{
			super("DataGraphException: " + e.getMessage());
		}
	}

	private Vector _data_nodes = null;  // nodes in the graph (hereafter referred to as data nodes)
	private Hashtable _id_data_node_map = null;  // mapping from ids to data nodes
	private DataNode _root_data_node = null;  // data node corresponding to the document root
	private Vector _out_edges_lists = null;  // edge lists

	private static int parserDebugLevel = Integer.getInteger("parserDebugLevel", 0).intValue();

	private static int nextDataNodeID = 0;

	// throws an exception if could not parse the file for any reason
	public DataGraph(String data_file_name_) throws DataGraphException
	{
		Node doc_root = parseFile(data_file_name_);
		constructGraph(doc_root);
	}

	public Vector getEdges(DataNode node)
	{ // returns Vector<DataNode>
		// find the list of outgoing edges for query_src_data_node_
		return (Vector)_out_edges_lists.elementAt(node.getGraphIndex());
	}

	// return the root node of the dataguide
	public DataNode getRoot()
	{
		return _root_data_node;
	}

	public DataNode getNode(int index)
	{
		return (DataNode) _data_nodes.elementAt(index);
	}

	// parse the file and return the dom node corresponding to the document root
	private Node parseFile(String data_file_name_) throws DataGraphException
	{
		Document document = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setValidating(false);  // Thomas: This does not seem to work
		factory.setNamespaceAware(true);
		// factory.setIgnoringComments(true); // Thomas
		// factory.setIgnoreElementContentWhitespace(true);
		// factory.setCoalescing(true);
		// factory.setExpandEntityReferences(false); // Thomas

		try
		{
			DocumentBuilder builder = factory.newDocumentBuilder();
//			document = builder.parse(new File(data_file_name_));
			if (parserDebugLevel > 1)
				System.out.println("Filename: " + data_file_name_);
			document = builder.parse(data_file_name_);
		}
		catch (SAXParseException spe)
		{
			// Error generated by the parser
/*			System.out.println ("\n** Parsing error"
								+ ", line " + spe.getLineNumber ()
								+ ", uri " + spe.getSystemId ());
			System.out.println("   " + spe.getMessage() );
			// Use the contained exception, if any
			Exception x = spe;
			if (spe.getException() != null)
				x = spe.getException();
			x.printStackTrace(); */
			throw new DataGraphException(spe);
		}
		catch (SAXException sxe)
		{
			// Error generated by this application
			// (or a parser-initialization error)
/*			Exception x = sxe;
			if (sxe.getException() != null)
				x = sxe.getException();
			x.printStackTrace(); */
			throw new DataGraphException(sxe);
		}
		catch (ParserConfigurationException pce)
		{
			// Parser with specified options can't be built
			// pce.printStackTrace();
			throw new DataGraphException(pce);
		}
		catch (IOException ioe)
		{
			// I/O error
			// ioe.printStackTrace();
			throw new DataGraphException(ioe);
		}

		Node doc_root = null;
		if (document != null)
			doc_root = document.getDocumentElement();  // init root to the root of the DOM structure
		return doc_root;
	} // private function parseFile

	private void constructGraph(Node doc_root_)
	{
		_data_nodes = new Vector();
		_out_edges_lists = new Vector();
		_id_data_node_map = new Hashtable();
		constructEltGraph(doc_root_);
		constructAttrGraph();
	}


	private void constructEltGraph(Node doc_root_)
	{
		Random random_num_gen = new Random();
		_root_data_node = constructEltGraph_helper(doc_root_, random_num_gen);
	}

	private DataNode constructEltGraph_helper(Node curr_node_, Random random_num_gen_)
	{
		// create id for current dom node if one is not specified in the file as the value of attrib. id
		NamedNodeMap attr_val_map = curr_node_.getAttributes();
		String id_string = null;
		// How about other forms of the string "id" --Thomas
		if (attr_val_map.getNamedItem("id") != null)
			id_string = attr_val_map.getNamedItem("id").getNodeValue().trim();
		if (id_string == null && attr_val_map.getNamedItem("ID") != null)
			id_string = attr_val_map.getNamedItem("ID").getNodeValue().trim();
		if (id_string == null)
			id_string = new String("Data." + random_num_gen_.nextInt());

		// keep track of idref attribute value for constructing cross (non-tree) edges later
		String idref_string = null;
		// How about other forms of the string "idref" --Thomas
		if (attr_val_map.getNamedItem("idref") != null)
			idref_string = attr_val_map.getNamedItem("idref").getNodeValue().trim();
		if (idref_string == null && attr_val_map.getNamedItem("IDREF") != null)
			idref_string = attr_val_map.getNamedItem("IDREF").getNodeValue().trim();

		String curr_node_name = curr_node_.getNodeName();
		String curr_node_data = null;
		NodeList sub_elements = curr_node_.getChildNodes();
		int num_sub_elements = sub_elements.getLength();
		for (int i = 0; i < num_sub_elements; i++)
		{
			Node sub_element = sub_elements.item(i);
			if (sub_element.getNodeType() == Node.TEXT_NODE ||
						sub_element.getNodeType() == Node.CDATA_SECTION_NODE)
			{
				curr_node_data = sub_element.getNodeValue();
				break;  // -- Thomas, How about merging disjoint pieces of text
			}
		}
		// create data node for the current dom node
		DataNode data_node = addDataNode(curr_node_name, curr_node_data);

		// adding non-id and non-idref attributes as nodes with incoming edges from the current data node
		int num_attrs = attr_val_map.getLength();
		for (int i = 0; i < num_attrs; i++)
		{
			Node attr_node = attr_val_map.item(i);
			if (attr_node.getNodeName().equalsIgnoreCase("id")) continue;
			if (attr_node.getNodeName().equalsIgnoreCase("idref")) continue;
//			String attr_id_string = new String("Data." + random_num_gen_.nextInt());
//			String attr_idref_string = null;
			String attr_node_name = attr_node.getNodeName();
			String attr_node_data = attr_node.getNodeValue();

			DataNode attr_data_node = addDataNode(attr_node_name, attr_node_data);
			addEdgeBetween(data_node, attr_data_node);
		}

		// add sub elements as nodes with incoming edges from the current data node
		NodeList sub_elements2 = curr_node_.getChildNodes();
		int num_sub_elements2 = sub_elements2.getLength();
		for (int i = 0; i < num_sub_elements2; i++)
		{
			Node sub_element = sub_elements2.item(i);
			if (sub_element.getNodeType() == Node.ELEMENT_NODE)
			{
				DataNode sub_element_data_node = this.constructEltGraph_helper(sub_element, random_num_gen_);
				addEdgeBetween(data_node, sub_element_data_node);
			}
		}
		return data_node;
	}

	private DataNode addDataNode(String nodeName, String nodeData)
	{
		DataNode theDataNode = new DataNode(nodeName, nodeData, _data_nodes.size());
		theDataNode.preorderIndex = nextDataNodeID++;
		_data_nodes.addElement(theDataNode);
		Vector out_edges_list = new Vector();
		_out_edges_lists.addElement(out_edges_list);
//		_id_data_node_map.put(theDataNode.getId(), theDataNode);
		return theDataNode;
	}

	// Add Non-Tree Edges
	private void constructAttrGraph()
	{
		// commented out because I'm only using it as a tree -- Donna

/*		int num_data_nodes = _data_nodes.size();
		for (int i = 0; i < num_data_nodes; i++)
		{
			// find data node that is target of idref and add an edge to it from the current data node
			DataNode curr_data_node = (DataNode)_data_nodes.elementAt(i);
			String idref_string = curr_data_node.getIdref();
			if (idref_string != null)
			{
				DataNode dest_data_node = (DataNode)_id_data_node_map.get(idref_string);
				// Tokenize idref string --Thomas
				// dest_data_node cannot be null --Thomas
				if (dest_data_node != null && !this.edgeExists(curr_data_node, dest_data_node))
					this.addEdgeBetween(curr_data_node, dest_data_node);
			}
		} */
	}

	private void addEdgeBetween(DataNode src_data_node_, DataNode dest_data_node_)
	{
		// find out_edge_list for src_data_node_
		boolean found = false;
		int num_data_nodes = _data_nodes.size();
		int index = 0;
		for (int i = 0; i < num_data_nodes; i++)
		{
			DataNode curr_data_node = (DataNode)_data_nodes.elementAt(i);
			if (curr_data_node == src_data_node_)
			{
				found = true; index = i; break;
			}
		}
		if (found == false)
		{
			System.out.println("data node not found in data graph");
		}
		Vector out_edges_list = (Vector)_out_edges_lists.elementAt(index);

		// add edge to dest_data_node_
		out_edges_list.addElement(dest_data_node_);
		dest_data_node_.setParentGraphIndex(index);
	}

	private boolean edgeExists(DataNode query_src_data_node_, DataNode query_dest_data_node_)
	{
		// find out_edge_list for src_data_node_
		boolean found = false;
		int num_data_nodes = _data_nodes.size();
		int index = 0;
		for (int i = 0; i < num_data_nodes; i++)
		{
			DataNode curr_data_node = (DataNode)_data_nodes.elementAt(i);
			if (curr_data_node == query_src_data_node_)
			{
				found = true; index = i; break;
			}
		}
		if (found == false)
		{
//			if (MyFrame._debug_mode > 6)
				System.out.println("data node not found in data graph");  // debug
			return false;
		}
		Vector out_edges_list = (Vector)_out_edges_lists.elementAt(index);

		// check if edge exists from query_src_data_node_ to query_dest_data_node_
		int num_out_edges = out_edges_list.size();
		for (int i = 0; i < num_out_edges; i++)
		{
			DataNode curr_data_node = (DataNode)out_edges_list.elementAt(i);
			if (query_dest_data_node_ == curr_data_node) return true;
		}
		return false;
	}
}

