/** DisplayNode.java
 *  Copyright (c) 2001 Donna Malayeri
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package edu.umd.cs.fuzzyTree;

import java.util.*;

/** Encapsulates a node displayed by a FuzzyTree
 *
 *  @author     Donna Malayeri
 *  @version    24-Jul-01
 */

public class DisplayNode
{
	/** Creates a DisplayNode that references <code>node</code> */
	public DisplayNode(DataNode node)
	{
		dataNode = node;
		dataNode.displayNode = this;
		numMissingChildren = FuzzyTreeModel.dataGraph.getEdges(dataNode).size();
	}

	/** The <code>DataNode</code> that is being displayed. Is never <code>null</code>. */
	public DataNode dataNode;

	/** A list of <code>DisplayNode</code>s that are the children of this node in the display tree.
	 *  This may not be the node's actual children.
	 *  @see #isOrphan
	 */
	public ArrayList children = new ArrayList();  // ArrayList<DisplayNode>

	/** The <code>DisplayNode</code> that is this node's parent */
	public DisplayNode parent;

	/** The number of nodes that are the child of the current node,
	 *  but are not its child in the display tree.
	 *  @see #isOrphan
	 *  @see #children
	 */
	public int numMissingChildren;

	/** Returns a string representation of this node, which can then be displayed in the
	 *  <code>FuzzyTree</code>.
	 *  If not debugging, only the node's label and data are included;
	 *  otherwise, its preorderIndex and score may also be included.
	 */
	public String toString()
	{
		String displayString = dataNode.getLabel() + "  " + dataNode.getData();

		if (FuzzyTreeModel.debugLevel >= 1)
			displayString += "       " + dataNode.preorderIndex;
		if (FuzzyTreeModel.debugLevel >= 2)
			displayString += "       " + dataNode.score;

		return displayString;

		// old code that uses HTML -- very slow
/*		String str = dataNode.getLabel() + "; " + dataNode.getData() +
			"&nbsp;&nbsp;&nbsp;&nbsp;" + dataNode.score +
			"<font color=#990000>&nbsp;&nbsp;&nbsp;&nbsp;" +
			dataNode.preorderIndex + "</font>";
		if (isOrphan())
			return "<html><font face=\"Arial, Helvetica, sans-serif\" size=2 color=blue>" + str + "</font>";
		else
			return "<html><font face=\"Arial, Helvetica, sans-serif\" size=2>" + str;
			return "<html><body bgcolor=\"#000000\"><font face=\"Arial, Helvetica, sans-serif\" size=2>" + str + "</body>"; */
	}

	/** Adds this node under parent <code>p</code>, in the correct position
	 *  (based on preorder index). Node is removed from the <code>children</code> list of its
	 *  previous parent, if it had one.
	 *  @param p    new display node parent for this node; must be non-null
	 */
	public void addUnderParent(DisplayNode p)
	{
		removeFromParent();
		parent = p;

		int childrenCursor;

		// find the index where this node should be inserted
		for (childrenCursor = 0; childrenCursor < p.children.size(); childrenCursor++)
		{
			DisplayNode node = (DisplayNode) p.children.get(childrenCursor);
			if (node.dataNode.preorderIndex > this.dataNode.preorderIndex)
				break;
		}

		p.children.add(childrenCursor, this);

		if (!isOrphan())
			parent.numMissingChildren--;
	}

	/** Removes this node from its parent's <code>children</code> list and sets the
	 *  <code>parent</code> reference to <code>null</code>.
	 */
	public void removeFromParent()
	{
		if (parent != null)
		{
			if (!parent.children.remove(this))
				throw new RuntimeException("Node was not a child of its parent: " + this);

			if (!isOrphan())
				parent.numMissingChildren++;
		}
	}

	/** Returns true if this node's actual parent is not its display parent.
	 *  @see FuzzyTreeModel#isOrphan
	 */
	public boolean isOrphan()
	{
		DataNode parentDataNode = dataNode.getParent(FuzzyTreeModel.dataGraph);
		return parent != null && parentDataNode != null && !parent.dataNode.equals(parentDataNode);
	}
}
