/** FuzzyTreeModel.java
 *  Copyright (c) 2001 Donna Malayeri
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package edu.umd.cs.fuzzyTree;

import javax.swing.tree.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;
import javax.swing.*;
import java.awt.*;

/**
 *  TreeModel used by FuzzyTree to display a partial tree of an XML document.  Each node
 *  in the tree is assigned a score which is a relative measure of how valuable the
 *  node is to be displayed. Expanding and collapsing subtrees are accomplished by adjusting
 *  the scores of nodes to add/remove them. Scores are assigned by the <code>ScoreProcessor</code>
 *  class.  The class has a variety of parameters which are used in the score increase/decrease
 *  algorithms, and which can be modified to produce different behavior.
 *
 *  @author     Donna Malayeri
 *  @version    24-Jul-01
 */

public class FuzzyTreeModel implements TreeModel
{
	/** Minimum factor of score to increase score when expanding.
	 *  <p>Default is 5% (.05) Default can be changed with the
	 *  environment variable <code>minExpandPercent</code>.
	 */
	private static float   minExpandFactor = loadPercentProperty("minExpandPercent", 5);

	/** Maximum factor of score to decrease score when collasping. Value is negative.
	 *  <p>Default is -5% (-.05) Default can be changed with the
	 *  environment variable <code>maxCollapsePercent</code>.
	 */
	private static float   maxCollapseFactor = loadPercentProperty("maxCollapsePercent", -5);

	/** Maximum factor of score to increase score when expanding.
	 *  <p>Default is 50% (.5) Default can be changed with the
	 *  environment variable <code>maxExpandPercent</code>.
	 */
	private static float   maxExpandFactor = loadPercentProperty("maxExpandPercent", 50);

	/** Maximum factor of score to decrease score when collasping. Value is negative.
	 *  <p>Default is -50% (-.50) Default can be changed with the
	 *  environment variable <code>minCollapsePercent</code>.
	 */
	private static float   minCollapseFactor = loadPercentProperty("minCollapsePercent", -50);

	/** Minimum number of children that should be displayed when user expands a subtree.
	 *  Score increase will be applied so that at least this many nodes are displayed, as long as
	 *  the factor increase is less than <code>maxExpandFactor</code>.
	 *  <p> Default value is 3.
	 *  Default can be changed with the environment variable <code>childrenToExpand</code>.*/
	private static int     minChildrenToExpand = Integer.getInteger("childrenToExpand", 3).intValue();

	/** Minimum number of children that should be removed when user collapses a subtree.
	 *  Score decrease will be applied so that at least this many nodes are removed, as long as
	 *  the factor decrease is greater than <code>minCollapseFactor</code>.
	 *  <p> Default value is 3.
	 *  Default can be changed with the environment variable <code>childrenToCollapse</code>.*/
	private static int     minChildrenToCollapse = Integer.getInteger("childrenToCollapse", 3).intValue();

	/** Factor by which to reduce score increases when moving down a subtree.
	 *  First level children will have score increase of <code>parentIncrease * increaseReductionPercent</code>,
	 *  and so on.
	 *  <p>Default value is 40%.  Default can be changed with the environment variable <code>increaseReductionPercent</code>.
	 */
	private static float   increaseReductionFactor = loadPercentProperty("increaseReductionPercent", 40);

	/** Factor by which to reduce score decreases when moving down a subtree.
	 *  First level children will have score decrease of <code>parentDecrease * decreaseReductionPercent</code>,
	 *  and so on.
	 *  <p>Default value is 40%.  Default can be changed with the environment variable <code>decreaseReductionPercent</code>.
	 */
	private static float   decreaseReductionFactor = loadPercentProperty("decreaseReductionPercent", 40);

	/** Threshold value at which score increases are no longer applied.
	 *  <p>Default value is 4%.
	 *  Default can be changed with the environment variable <code>increaseThresholdPercent</code>.
	 */
	private static float   increaseThresholdValue = loadPercentProperty("increaseThresholdPercent", 4);

	/** Threshold value at which score decreases are no longer applied.
	 *  <p>Default value is -4%.
	 *  Default can be changed with the environment variable <code>decreaseThresholdPercent</code>.
	 */
	private static float   decreaseThresholdValue = loadPercentProperty("decreaseThresholdPercent", -4);

	/** Number of user history entries to maintain.  When the user history is full, the oldest user
	 *  history item is undone. Default value is 10, can be changed with the environment variable
	 *  <code>userHistorySize</code>.
	 *  @see #_userActions
	 */
	private int     _userHistorySize = Integer.getInteger("userHistorySize", 10).intValue();

	/** Reference to the XML data graph. */
	static DataGraph dataGraph;

	/** For creating initial sorted list from the DataGraph. */
	private ScoreGenerator _scoreGenerator;

	/** A sorted list of all of the <code>DataNode</code>s in the DataGraph.
	 *  Nodes are sorted based on the DataNode's score field.
	 */
	private ArrayList _sortedNodeList; // ArrayList<DataNode>

	/** List of user actions.  Size of list is <code>_userHistorySize</code>
	 * <p><code>ArrayList</code> of <code>ArrayList</code> of <code>UserAction</code>
	 */
	private ArrayList _userActions = new ArrayList(_userHistorySize); 	// ArrayList<ArrayList<UserAction> >

	/** List of temporary actions.  Temporary actions are score changes that last only until the next
	 *  action.  Examples: increasing the score of a node that was expanded or collapsed to ensure
	 *  that it will be visible for at least one action
	 *  <p><code>ArrayList</code> of <code>UserAction</code>.
	 */
	private ArrayList _temporaryUserActionList = new ArrayList(); // ArrayList<UserAction>

	/** Number of nodes currently being displayed in addition to the root.
	 *  (<code>_numDisplayNodes + 1</code> nodes are always displayed)
	 *  Setter and getter methods do the +/- 1 adjustment.
	 *  Nodes from index 0 to <code>_numDisplayedNodes - 1</code> of
	 *   <code>_sortedNodeList</code> will be displayed
	 */
	private int _numDisplayedNodes;

	/** Root of the displayed tree. Is always displayed, so is not in <code>_sortedNodeList</code>. */
	private DisplayNode _treeRoot;
	/** Node that was last clicked on, which is highlighted. */
	private DataNode _activeNode;

	/** Debug output; output to System.out, unless the environment variable debugFilename is specified */
	private PrintWriter _debugOutput;
	/** Debug output; output to System.out, unless the enviroment variable actionsFilename is specified.
	 *  If debug level is set, all user actions are outputed to this. */
	private PrintWriter _actionsOutput;

	/** Debug levels (each level includes actions of lower levels).
	 *  <ul>
	 *  <li> Level 1: Print actions only, display preorder index of each node in tree <br>
	 *  <li> Level 2: Show scores of each node in tree <br>
	 *  <li> Level 3: "Sanity check" - ensure that only <code>numDisplayedNodes - 1</code>
	 *      of <code>_sortedNodeList</code> are displayed<br>
	 *  <li> Level 4: Print score changes <br>
	 *  <li> Level 5: Print whole lists <br>
	 *  <li> Level 6: Print initial sorted list <br>
	 *  </ul>
	 */
	public static int debugLevel = Integer.getInteger("debugLevel", 0).intValue();

	/** Simple data structure, consists of affected node and an associated score change. */
	private static class UserAction
	{
		public UserAction(DataNode node)
		{
			actionNode = node;
		}

		public UserAction(DataNode node, int sc)
		{
			actionNode = node;
			scoreChange = sc;
		}

		public DataNode actionNode;
		public int scoreChange;
	}

	/** Listeners of TreeModel events. */
	EventListenerList _listeners = new EventListenerList();

	/** For sorting in ascending order of preorder index. */
	Comparator _preorderIndexComparator =
		new Comparator() {
			public int compare(Object o1, Object o2)
			{
				return ((DisplayNode)o1).dataNode.preorderIndex -
						((DisplayNode)o2).dataNode.preorderIndex;
			}
		};

	/** Create FuzzyTreeModel from the XML file passed in.
	 *  @param xmlFilename
	 *      XML file to open and create tree from. File <b>must</b> be a tree (not a graph)
	 *  @throws DataGraph.DataGraphException    if there was an error opening/parsing the file
	 */
	public FuzzyTreeModel(String xmlFilename) throws DataGraph.DataGraphException
	{
		_debugOutput = createDebugOutput("debugFilename", false);
		_actionsOutput = createDebugOutput("actionsFilename", true);

		dataGraph = new DataGraph(xmlFilename);
		_treeRoot = new DisplayNode((DataNode) dataGraph.getRoot());

		_scoreGenerator = new ScoreGenerator(dataGraph);
		_sortedNodeList = _scoreGenerator.getSortedNodeList();

		// debug output
		if (debugLevel >= 6)
		{
			Iterator iter = _sortedNodeList.iterator();
			while (iter.hasNext())
			{
				DataNode item = (DataNode) iter.next();
				_debugOutput.println(item);
			}

			_debugOutput.flush();
		}
	}

	// **** start TreeModel methods ****

	public Object getRoot()
	{
		return _treeRoot;
	}

	public Object getChild(Object parent, int index)
	{
		return ((DisplayNode) parent).children.get(index);
	}

	public int getChildCount(Object parent)
	{
		return ((DisplayNode) parent).children.size();
	}

	public boolean isLeaf(Object node)
	{
		// a node is a leaf if it's not the root and if it has no children in the datagraph
		return !node.equals(_treeRoot) &&
			dataGraph.getEdges(((DisplayNode)node).dataNode).size() == 0;
	}

	/** TreeModel method.  Not implemented. */
	public void valueForPathChanged(TreePath path, Object newValue)
	{
	}

	public int getIndexOfChild(Object parent, Object child)
	{
		return ((DisplayNode) parent).children.indexOf(child);
	}

	public void addTreeModelListener(TreeModelListener l)
	{
		_listeners.add(TreeModelListener.class, l);
	}

	public void removeTreeModelListener(TreeModelListener l)
	{
		_listeners.remove(TreeModelListener.class, l);
	}

	// **** end TreeModel methods ****

	// ***** TreeModel properties *****

	/** Returns true if <code>node</code> should be highlighted.
	 *  @see #_activeNode
	 */
	public boolean isActiveNode(Object node)
	{
		return ((DisplayNode)node).dataNode.equals(_activeNode);
	}

	/** Returns total number of nodes in the tree. */
	public int getTotalNodes()
	{
		return 1 + _sortedNodeList.size();
	}

	/** Get parameters used in score adjusting algorithms
	 *  @return parameters used by score adjusting algorithms */
	public FuzzyTreeParameters getParameters()
	{
		return new FuzzyTreeParameters(minExpandFactor * 100, maxExpandFactor * 100,
						maxCollapseFactor * 100, minCollapseFactor * 100,
						increaseReductionFactor * 100, increaseThresholdValue * 100,
						decreaseReductionFactor * 100, decreaseThresholdValue * 100,
						minChildrenToExpand, _userHistorySize);
	}

	/** Set parameters for score adjusting algorithms
	 *  @param newParameters    new parameter values
	 *  */
	public void setParameters(FuzzyTreeParameters newParameters)
	{
		minExpandFactor = newParameters.minExpandPercent / 100;
		maxExpandFactor = newParameters.maxExpandPercent/ 100;
		maxCollapseFactor = newParameters.maxCollapsePercent / 100;
		minCollapseFactor = newParameters.minCollapsePercent / 100;
		increaseReductionFactor = newParameters.percentIncreaseReduction / 100;
		increaseThresholdValue = newParameters.percentIncreaseThreshold / 100;
		decreaseReductionFactor = newParameters.percentDecreaseReduction / 100;
		decreaseThresholdValue = newParameters.percentDecreaseThreshold / 100;
		minChildrenToExpand = newParameters.minChildrenToExpand;
		minChildrenToCollapse = newParameters.minChildrenToExpand;

		if (_userHistorySize != newParameters.userHistorySize)
		{
			for (int i = _userActions.size() - 1; i >= 0; i--)
				undoAction((ArrayList)_userActions.remove(i));

			_userHistorySize = newParameters.userHistorySize;
			_userActions.ensureCapacity(_userHistorySize);
		}
	}

	/** Returns number of nodes that are displayed
	 *  @return number of displayed nodes, including the root
	 */
	public int getNumDisplayedNodes()
	{
		return _numDisplayedNodes + 1;  // adjust for root
	}

	/** Set new value for number of displayed nodes
	 *  @param newValue     new number of nodes to display
	 */
	public void setNumDisplayedNodes(int newValue)
	{
		newValue--; // adjust for root

		if (_numDisplayedNodes == newValue)
			return;

		if (newValue > _sortedNodeList.size())
			throw new RuntimeException("setNumDisplayedNodes was called with: " +
				(newValue + 1) + " total nodes: " + _sortedNodeList.size());

		undoTemporaryAction();

		int currentNodesDisplayed = _numDisplayedNodes;
		_numDisplayedNodes = newValue;

		if (debugLevel >= 1)
			_actionsOutput.println((_numDisplayedNodes + 1) + " nodes displayed");

		if (currentNodesDisplayed < _numDisplayedNodes)
			addDisplayNodeRange(currentNodesDisplayed);
		else
			removeDisplayNodeRange(currentNodesDisplayed);

		fireTreeStructureChanged();
	}

	// ***** End tree model properties ****

	// ***** Static methods ****

	/** Returns true if <code>node</code> is not a child of its parent,
	 *  but rather a child of an ancestor.
	 *  @see DisplayNode#isOrphan */
	public static boolean isOrphan(Object node)
	{
		return ((DisplayNode)node).isOrphan();
	}

	/** Node is fully open (all of its children are displayed). */
	public static final int NODE_FULLY_OPEN = 0;
	/** Node is closed (none of its children are displayed */
	public static final int NODE_CLOSED = 1;
	/** Node is partially open (some of its children are displayed). */
	public static final int NODE_PARTIALLY_OPEN = 2;

	/** Returns the open/close state of <code>node</code>.
	 *  @return     One of <code>NODE_FULLY_OPEN, NODE_CLOSED, NODE_PARTIALLY_OPEN</code>
	 */
	public static int getNodeState(Object node)
	{
		DisplayNode displayNode = (DisplayNode) node;

		if (displayNode.children.isEmpty())
			return NODE_CLOSED;
		else if (displayNode.numMissingChildren == 0)
			return NODE_FULLY_OPEN;
		else
			return NODE_PARTIALLY_OPEN;
	}

	// **** Public Methods *****

	/** Tries to expand the subtree that <code>node</code> is the root of.  Expansion is done through
	 *  score changes and amount of score increase is constrained by
	 *  <code>minExpandFactor</code> and <code>maxExpandFactor</code>.
	 *  Score change to display <code>childrenToExpand</code> nodes is applied if <code>fullyExpand</code>
	 *  is </code>false</code>, otherwise, score change to display all children is applied.
	 *  <p>Applies score changes to descendants further down the tree, but the score
	 *  increase is reduced at each level by <code>increaseReductionFactor</code>.
	 *  Score change is not applied if it is less than <code>increaseThresholdValue</code>.
	 */
	public void userExpandSubtree(Object node, boolean fullyExpand)
	{
		DisplayNode displayNode = (DisplayNode) node;
		DataNode sourceNode = displayNode.dataNode;
		_activeNode = sourceNode;

		if (debugLevel > 1)
		{
			System.out.println( (fullyExpand ? "++\t" : "+\t") + sourceNode );
			_debugOutput.println( (fullyExpand ? "++\t" : "+\t") + sourceNode );
			_actionsOutput.println( (fullyExpand ? "++\t" : "+\t") + sourceNode );
		}

		int numAddedNodes = 0;  // the number of nodes that have been added to the display tree
								// as a result of the expansion

		ArrayList actionList = new ArrayList();
		boolean undoWasDone = addUserAction(actionList);  // this will undo old actions as well

		// inflate the score of the node that was clicked on to be the same as the highest
		// scoring node, temporarily
		UserAction temporaryUserAction = new UserAction(sourceNode);
		if (inflateNodeScore(sourceNode, temporaryUserAction))
			numAddedNodes++;

		// undo or score inflation may have changed displayNode. set it to the correct value
		displayNode = sourceNode.displayNode;

		Vector dataNodeChildren = dataGraph.getEdges(sourceNode);
		// assert(!dataNodeChildren.isEmpty())

		// make a copy of the children and sort in descending order on score
		Vector childrenCopy = (Vector) dataNodeChildren.clone();
		Collections.sort(childrenCopy, _scoreGenerator.descendingScoreComparator);

		// determine what factor increase should be applied
		// it will be the factor required to display either MIN_NODES_TO_EXPAND or
		// all of the children, depending on the value of fullyExpand
		int nodesToExpand = fullyExpand ? dataNodeChildren.size() : minChildrenToExpand;
		int lastIndex = nodesToExpand - 1;
		if (lastIndex >= childrenCopy.size());
			lastIndex = childrenCopy.size() - 1;
		DataNode lastChild = (DataNode) childrenCopy.elementAt(lastIndex);
		int lastSortedNodeIndex = _numDisplayedNodes - nodesToExpand;
		if (lastSortedNodeIndex >= _sortedNodeList.size())
			lastSortedNodeIndex = _sortedNodeList.size() - 1;
		if (lastSortedNodeIndex < 0)
			lastSortedNodeIndex = 0;

		int lowestDisplayedScore = ((DataNode)_sortedNodeList.get(lastSortedNodeIndex)).score;
		float mainIncreaseFactor =  (float) (lowestDisplayedScore - lastChild.score) / lastChild.score;
		float factorIncreaseToDisplaySourceNode = (float) (lowestDisplayedScore - sourceNode.score) / sourceNode.score;

		debugPrint(4, "\n\nPercent increase: " + mainIncreaseFactor);

		// the percent increase should be between the minimum and the maximum
		mainIncreaseFactor = Math.max(mainIncreaseFactor, minExpandFactor);
		mainIncreaseFactor = Math.min(mainIncreaseFactor, maxExpandFactor);

		debugPrintln(4, " --> " + mainIncreaseFactor);

		// update the score of the expanded node by same amount as children
		numAddedNodes += increaseNodeScore(sourceNode, mainIncreaseFactor, actionList);
		displayNode = sourceNode.displayNode;   // this may be different than display node,
												// if an undo was done and the node was added
												// back to the tree

		// update scores of childen, going all the way down the tree
		Iterator iter = dataNodeChildren.iterator();
		while (iter.hasNext())
		{
			numAddedNodes +=
				increaseSubtreeScore((DataNode) iter.next(), mainIncreaseFactor, actionList);
		}

		// Go up the tree and find all the parents whose scores should change
		float curFactorIncrease = mainIncreaseFactor;
		DisplayNode ancestor = displayNode;

		if (displayNode != null && ancestor.parent != null)
		{
			while ( (ancestor = ancestor.parent) != _treeRoot &&
					curFactorIncrease > increaseThresholdValue)
			{
				// reduce increase factor to some percentage of before
				curFactorIncrease *= increaseReductionFactor;
				UserAction action = new UserAction(ancestor.dataNode);
				action.scoreChange = (int) Math.ceil(ancestor.dataNode.score * curFactorIncrease);
				actionList.add(action);

				debugPrint(4, "Updated score of " + ancestor.dataNode);
				ancestor.dataNode.score += action.scoreChange;
				debugPrintln(4, "  to " + ancestor.dataNode.score);

				repositionNodeInSortedList(ancestor.dataNode, INSERT_NODE_BEFORE);
				// don't need to add the ancestor, since it is necessarily already in the tree
			}
		}

		debugPrintln(4, numAddedNodes + " nodes added total");
		debugFlush(4);

		debugPrintln(5, "*** The new list (before removal)");
		debugPrintln(5, "");
		printDisplayedNodes(_debugOutput, _numDisplayedNodes + numAddedNodes);

		// since numAddedNodes got added, that means that many got bumped off. remove them
		int nodeListCursor = _numDisplayedNodes;
		int numRemovedNodes = 0;

		while (numRemovedNodes < numAddedNodes)
		{
			DataNode bumpedOffNode = (DataNode) _sortedNodeList.get(nodeListCursor);

			if (bumpedOffNode.displayNode != null)
			{
				numRemovedNodes++;
				debugPrintln(4, "Bumping off " + bumpedOffNode + " " + numRemovedNodes);
				removeDisplayNode(bumpedOffNode);
			}

			nodeListCursor++;
		}

		printDisplayedNodes(_debugOutput);

		// tree has been changed, notify listeners
		fireTreeStructureChanged();
	}

	/** Tries to collapse the subtree that <code>node</code> is the root of.
	 *  Collapsing is done through score changes and amount of score decrease is constrained by
	 *  <code>minCollapseFactor</code> and <code>maxCollapseFactor</code>.
	 *  Applies score change to remove <code>childrenToCollapse</code> nodes if
	 *  <code>fullyExpand</code> is </code>false</code>,
	 *  otherwise, applies score change to remove all children.
	 *  <p>Applies score changes to descendants further down the tree, but the score
	 *  decrease will is reduced at each level by <code>decreaseReductionFactor</code>.
	 *  Score change is not applied if it is less than <code>decreaseThresholdValue</code>.
	 */
	public void userCollapseSubtree(Object node, boolean fullyCollapse)
	{
		DisplayNode displayNode = (DisplayNode) node;
		DataNode sourceNode = displayNode.dataNode;
		_activeNode = sourceNode;

		if (debugLevel >= 1)
		{
			System.out.println( (fullyCollapse ? "--\t" : "-\t") + sourceNode );
			_debugOutput.println();
			_debugOutput.println( (fullyCollapse ? "--\t" : "-\t") + sourceNode );
			_debugOutput.println();
			_actionsOutput.println( (fullyCollapse ? "--\t" : "-\t") + sourceNode );
		}

		int numRemovedNodes = 0;    // the number of nodes that have been removed as a result
									// of the collapse

		ArrayList actionList = new ArrayList();
		addUserAction(actionList);  // this will undo old actions as well

		// inflate the score of the node that was clicked on to be the same as the highest
		// scoring node, temporarily
		int oldSourceScore = sourceNode.score;

		UserAction temporaryUserAction = new UserAction(sourceNode);
		boolean sourceNodeWasAdded = inflateNodeScore(sourceNode, temporaryUserAction);

		if (sourceNodeWasAdded)
			numRemovedNodes--;

		// undo, or score inflation may have changed displayNode
		displayNode = sourceNode.displayNode;

		// determine what decrease factor should be applied
		// it will be the value required to collapse either MIN_NODES_TO_EXPAND or
		// all of the children, depending on the value of fullyCollapse

		// make a copy of the displayed children and sort in ascending order on score
		ArrayList childrenCopy = (ArrayList) displayNode.children.clone();

		Collections.sort(childrenCopy, new Comparator()
			{
				public int compare(Object o1, Object o2)
				{
					return ((DisplayNode)o1).dataNode.score - ((DisplayNode)o2).dataNode.score;
				}
			});

		float decreaseFactor = 0;

		if (!childrenCopy.isEmpty())
		{
			int nodesToCollapse = fullyCollapse ? childrenCopy.size() : minChildrenToCollapse;
			int lastIndex = nodesToCollapse - 1;
			if (lastIndex >= childrenCopy.size());
				lastIndex = childrenCopy.size() - 1;
			DataNode lastChild = ((DisplayNode) childrenCopy.get(lastIndex)).dataNode;
			int lastSortedNodeIndex = _numDisplayedNodes + nodesToCollapse;
			if (lastSortedNodeIndex >= _sortedNodeList.size())
				lastSortedNodeIndex = _sortedNodeList.size() - 1;
			decreaseFactor =  (float) (((DataNode)_sortedNodeList.get(lastSortedNodeIndex)).score -
				lastChild.score) / lastChild.score;

			decreaseFactor = Math.min(decreaseFactor, maxCollapseFactor);
			decreaseFactor = Math.max(decreaseFactor, minCollapseFactor);

			// apply the decrease factor directly to the nodes that are displayed
			// (this is slightly different from the case for expanding)
			Iterator iter = childrenCopy.iterator();
			while (iter.hasNext())
			{
				DisplayNode child = (DisplayNode) iter.next();
				numRemovedNodes += decreaseNodeScore(child.dataNode, decreaseFactor, actionList);
			}
		}

		// update scores of childen, going all the way down the tree
		// encountered that are nodes already in the display tree will not have their scores updated
		Vector dataNodeChildren = dataGraph.getEdges(sourceNode);
		Iterator iter = dataNodeChildren.iterator();
		while (iter.hasNext())
		{
			numRemovedNodes +=
				decreaseSubtreeScore((DataNode) iter.next(), decreaseFactor, actionList);
		}

		// determine by how much the score of the clicked-on node should be decremented,
		// at the next action
		int scoreDecrease = (int) Math.floor(oldSourceScore * decreaseFactor);
		actionList.add(new UserAction(sourceNode, scoreDecrease));
		temporaryUserAction.scoreChange += (-scoreDecrease);
		debugPrintln(4, "Total temporary score change: " + temporaryUserAction.scoreChange +
						   " (decrement was " + scoreDecrease + ")");

		debugPrintln(4, "");
		debugPrintln(4, numRemovedNodes + " nodes removed");

//		int nodeListCursor = _numDisplayedNodes - numRemovedNodes;  // wrong!!

		if (numRemovedNodes < 0)
		{
			debugPrintln(4, "sourceNodeWasAdded, removing a node");

			int nodeListCursor = _numDisplayedNodes;
			boolean removed = false;

			while (!removed)
			{
				DataNode bumpedOffNode = (DataNode) _sortedNodeList.get(nodeListCursor);

				if (bumpedOffNode.displayNode != null)
				{
					removed = true;
					debugPrintln(4, "Bumping off " + bumpedOffNode + " " + numRemovedNodes);
					removeDisplayNode(bumpedOffNode);
				}

				nodeListCursor++;
			}
		}

		// since numRemovedNodes got removed, that means that many got added.
		// add them to the display tree
		else
		{
			int nodeListCursor = 0;
			int numAddedNodes = 0;

			while (numAddedNodes < numRemovedNodes)
			{
				DataNode addNode = (DataNode) _sortedNodeList.get(nodeListCursor);

				if (addNode.displayNode == null)
				{
					numAddedNodes++;
					debugPrintln(4, "Adding " + addNode);
					addDisplayNode(addNode);
				}

				nodeListCursor++;
			}
		}

		printDisplayedNodes(_debugOutput);
		fireTreeStructureChanged();

	}

	protected void fireTreeStructureChanged()
	{
		if (debugLevel >= 3)
			doSanityCheck();

		// tell listeners about the tree
		Object[] listeners = _listeners.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i]==TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e = new TreeModelEvent(this, new Object[] { _treeRoot });
				((TreeModelListener)listeners[i+1]).treeStructureChanged(e);
			}
		}
	}

	/** Increases the score of <code>dataNode</code> by <code>increaseFactor</code>.
	 *  Adds action of updating the score to <code>actionList</code>.
	 *
	 *  @param dataNode         the node to which to apply the score change
	 *  @param increaseFactor   factor score increase
	 *  @param actionList       list to add the action of score increase to
	 *
	 *  @return     the number of nodes that were added to the display tree (0 or 1)
	 */
	private int increaseNodeScore(DataNode dataNode, float increaseFactor,
									ArrayList actionList)
	{
		int numAddedNodes = 0;

		String debugString =  "Updated score of " + dataNode;
		int scoreIncrease = (int) Math.ceil(dataNode.score * increaseFactor);
		dataNode.score += scoreIncrease;

		debugString += " to " + dataNode.score;

		actionList.add(new UserAction(dataNode, scoreIncrease));

		int newIndex = repositionNodeInSortedList(dataNode, INSERT_NODE_BEFORE);

		// add the node if it wasn't in the tree before and it should be now
		if (newIndex < _numDisplayedNodes && dataNode.displayNode == null)
		{
			addDisplayNode(dataNode);
			numAddedNodes++;

			debugString = "* " + debugString;
		}

		debugPrintln(4, debugString);

		return numAddedNodes;
	}

	/** Increases the scores in the subtree of which
	 *  <code>dataNode</code> is the root, by <code>increaseFactor</code>.
	 *  Reduces the amount by which scores are increased at each level by
	 *  <code>increaseReductionFactor</code>.
	 *  Adds the action of updating the score to <code>actionList</code>.
	 *  <p>Method is called recursively.
	 *
	 *  @param dataNode         the node to which to apply the score change
	 *  @param increaseFactor   factor score increase
	 *  @param actionList       list to add the action of score increase to
	 *
	 *  @return     the number of nodes that were added to the display tree
	 */
	private int increaseSubtreeScore(DataNode dataNode, float increaseFactor,
									ArrayList actionList)
	{
		int numAddedNodes = increaseNodeScore(dataNode, increaseFactor, actionList);

		// update scores of children, but by a smaller increase factor
		increaseFactor *= increaseReductionFactor;
		if (increaseFactor > increaseThresholdValue)
		{
			Iterator iter = dataGraph.getEdges(dataNode).iterator();
			while (iter.hasNext())
			{
				numAddedNodes += increaseSubtreeScore((DataNode) iter.next(),
													increaseFactor, actionList);
			}
		}

		return numAddedNodes;
	}

	/** Reduces the score of <code>dataNode</code> by <code>decreaseFactor</code>.
	 *  Adds action of updating the score to <code>actionList</code>.
	 *
	 *  @param dataNode         the node to which to apply the score change
	 *  @param decreaseFactor   factor score decrease
	 *  @param actionList       list to add the action of score decrease to
	 *
	 *  @return     the number of nodes that were removed from the display tree (0 or 1)
	 */
	private int decreaseNodeScore(DataNode dataNode, float decreaseFactor, ArrayList actionList)
	{
		int numRemovedNodes = 0;

		String debugString =  "Decreased score of " + dataNode;
		int scoreDecrease = (int) Math.floor(dataNode.score * decreaseFactor);
		dataNode.score += scoreDecrease;

		debugString += " to " + dataNode.score;

		actionList.add(new UserAction(dataNode, scoreDecrease));

		int newIndex = repositionNodeInSortedList(dataNode, INSERT_NODE_AFTER);

		// remove the node if it needs to be removed
		if (newIndex >= _numDisplayedNodes && dataNode.displayNode != null)
		{
			removeDisplayNode(dataNode);
			numRemovedNodes++;

			debugString = "~ " + debugString;
		}

		debugPrintln(4, debugString);

		return numRemovedNodes;
	}

	/** Decreases the scores in the subtree of which
	 *  <code>dataNode</code> is the root, by <code>decreaseFactor</code>.
	 *  Reduces the amount by which scores are decreased at each level by
	 *  <code>decreaseReductionFactor</code>.
	 *  Adds the action of updating the score to <code>actionList</code>.
	 *  <p>Does not apply score change to nodes that are in the display tree.
	 *  <p>Method is called recursively.
	 *
	 *  @param dataNode         the node to which to apply the score change
	 *  @param decreaseFactor   factor score decrease
	 *  @param actionList       list to add the action of score decrease to
	 *
	 *  @return     the number of nodes that were removed from the display tree
	 */
	private int decreaseSubtreeScore(DataNode dataNode, float decreaseFactor,
									ArrayList actionList)
	{
		int numRemovedNodes = 0;

		if (dataNode.displayNode == null)
			numRemovedNodes = decreaseNodeScore(dataNode, decreaseFactor, actionList);

		// reduce scores of children, but by a smaller decrease factor
		decreaseFactor *= decreaseReductionFactor;
		if (decreaseFactor > decreaseThresholdValue)
		{
			Iterator iter = dataGraph.getEdges(dataNode).iterator();
			while (iter.hasNext())
			{
				numRemovedNodes += decreaseSubtreeScore((DataNode) iter.next(),
													  decreaseFactor, actionList);
			}
		}

		return numRemovedNodes;
	}

	/** Inflates the score of <code>sourceNode</code> such that it will be the highest-scoring
	 *  node in the tree.  Sets the score change amount in <code>temporaryUserAction</code>.
	 *
	 *  @param sourceNode               node to which to apply the score change
	 *  @param temporaryUserAction      user action to record the score change.
	 *      <code>actionNode</code> field should already be set
	 *
	 *  @return     <code>true</code> if node was added to the tree as a result of score inflation;
	 *              <code>false</code> otherwise
	 */
	private boolean inflateNodeScore(DataNode sourceNode, UserAction temporaryUserAction)
	{
		int inflateAmount = ((DataNode) _sortedNodeList.get(0)).score - sourceNode.score;
		temporaryUserAction.scoreChange = inflateAmount;
		_temporaryUserActionList.add(temporaryUserAction);
		debugPrintln(4, "Inflated score of " + sourceNode + " to " + (inflateAmount + sourceNode.score));
		sourceNode.score += inflateAmount;

		int newIndex = repositionNodeInSortedList(sourceNode, INSERT_NODE_BEFORE);

		// add the node if it wasn't in the tree before and it should be now
		if (newIndex < _numDisplayedNodes && sourceNode.displayNode == null)
		{
			debugPrintln(4, "Node now displayed as result of score inflation: " + sourceNode);
			addDisplayNode(sourceNode);
			return true;
		}
		else
			return false;
	}

	/** Undoes the score change associated with the last temporary action.
	 *  @see #_temporaryUserActionList
	 */
	private void undoTemporaryAction()
	{
		if (!_temporaryUserActionList.isEmpty())
		{
			undoAction(_temporaryUserActionList);
			_temporaryUserActionList.clear();
		}
	}

	/** Adds a new set of user actions to the user history.  Undoes old actions if the list is full.
	 *  This method should be called before score change values are calculated, as an undo
	 *  will change the state of the tree.
	 *
	 *  @return     true if adding the action list resulted in another action being undone
	 */
	private boolean addUserAction(ArrayList actionList) 	// ArrayList<UserAction> actionList
	{
		// undo last temporary action
		undoTemporaryAction();

		_userActions.add(actionList);

		if (_userActions.size() == _userHistorySize + 1)
		{
			debugPrintln(4, "\n#### Undoing first action");

			// remove the first action and undo it
			undoAction((ArrayList) _userActions.remove(0));
			debugPrintln(5, "\nAfter undo:\n");
			printDisplayedNodes(_debugOutput);

			doSanityCheck();

			return true;
		}

		else
		{
			doSanityCheck();
			return false;
		}
	}

	/** Iterates through list of user actions and reverses the score change associated with each one.
	 *  Nodes are added/removed from the display tree as necessary
	 *
	 *  @param  actions     list of <code>UserAction</code>s to undo
	 */
	private void undoAction(ArrayList actions)  // ArrayList<UserAction> actions
	{
		// a more efficient way to implement this method might be to keep a counter of
		// nodes to add/remove, then at the end, add or remove as necessary

		Iterator iter = actions.iterator();
		while (iter.hasNext())
		{
			UserAction action = (UserAction) iter.next();

			// undo the score change to the affected node
			action.actionNode.score -= action.scoreChange;

			// make sure the scores is at least the minimum node score
			action.actionNode.score = Math.max(action.actionNode.score,
										_scoreGenerator.MIN_NODE_SCORE);

			int newIndex =
				repositionNodeInSortedList(action.actionNode,
					action.scoreChange < 0 ? INSERT_NODE_BEFORE : INSERT_NODE_AFTER);

			if (action.actionNode.displayNode != null && newIndex >= _numDisplayedNodes)
			{
				removeDisplayNode(action.actionNode);

				// this means that the node at _numDisplayedNodes - 1 must necessarily be added
				DataNode addNode = (DataNode) _sortedNodeList.get(_numDisplayedNodes - 1);
				addDisplayNode(addNode);

//				System.out.println("- Removing " + action.actionNode);
//				System.out.println("^ Adding " + addNode);
			}
			else if (action.actionNode.displayNode == null && newIndex < _numDisplayedNodes)
			{
				addDisplayNode(action.actionNode);
				// this means that the node at _numDisplayedNodes must necessarily be removed
				DataNode removeNode = (DataNode) _sortedNodeList.get(_numDisplayedNodes);
				removeDisplayNode(removeNode);

//				System.out.println("^ Adding " + action.actionNode);
//				System.out.println("- Removing " + removeNode);
			}
			else
			{
//				System.out.println("Updated score, no change " + action.actionNode);
			}
		}
	}

	/** Adds <code>_numDisplayedNodes - currentNumNodes</code> nodes to the display tree.
	 *
	 *  @param currentNumNodes      the number of nodes already displayed
	 */
	private void addDisplayNodeRange(int currentNumNodes)
	{
		for (int i = currentNumNodes; i < _numDisplayedNodes; i++)
		{
			// get ith scoring member
			DataNode node = (DataNode) _sortedNodeList.get(i);
			addDisplayNode(node);
		}
	}

	/** Adds <code>node</code> to the display tree.  It is positioned under any parent already
	 *  in the tree. If its parent is not in the tree, it is positioned under its closest ancestor,
	 *  and it will be an orphan.
	 *  Any descendants the node has that are in the display tree are positioned under it.
	 *
	 *  @param   node        node to add to the display tree
	 */
	private void addDisplayNode(DataNode node)
	{
		DisplayNode displayNode = new DisplayNode(node);

		DataNode parent = node;
		boolean foundParent = false;

		while ( !foundParent && (parent = parent.getParent(dataGraph)) != null)
		{
			if (parent.displayNode != null)
			{
				displayNode.addUnderParent(parent.displayNode);
				foundParent = true;
			}
		}

		if (!foundParent)
			displayNode.addUnderParent(_treeRoot);

		// check to see if any children or grandchildren, etc, are in the tree, if so,
		// make them children of this node
		Iterator childIter = dataGraph.getEdges(node).iterator();
		while (childIter.hasNext())
		{
			DataNode child = (DataNode) childIter.next();
			moveSubtree(child, displayNode);
		}
	}

	/** Removes <code>currentDisplayedNodes - _numDisplayedNodes</code> nodes from
	 *  the displayTree.
	 *
	 *  @param currentNumNodes      the number of nodes currently displayed
	 */
	private void removeDisplayNodeRange(int currentDisplayedNodes)
	{
		for (int i = _numDisplayedNodes; i < currentDisplayedNodes; i++)
		{
			// get ith scoring member
			DataNode node = (DataNode) _sortedNodeList.get(i);
			removeDisplayNode(node);
		}
	}

	/** Removes <code>node</code> from the display tree.  Children become the children of
	 *  <code>node</code>'s parent.
	 *
	 *  @param   node        node to remove from the display tree
	 */
	private void removeDisplayNode(DataNode node)
	{
		DisplayNode deleteNode = node.displayNode;

		deleteNode.removeFromParent();

		// move children of deleted node to under deleteNode's parent
		if (!deleteNode.children.isEmpty())
		{
			Iterator iter = deleteNode.children.iterator();
			while (iter.hasNext())
			{
				DisplayNode child = (DisplayNode) iter.next();
				child.parent = null;    // reset the parent pointer
				child.addUnderParent(deleteNode.parent);
			}
		}

		node.displayNode = null;    // node is no longer in displayed tree
		deleteNode.parent = null;
		deleteNode.children = null;
	}

	/** Moves <code>child</code> under <code>displayParent</code>,
	 *  if it is in the displayed tree; otherwise, any displayed descendants of <code>child</code>
	 *  are moved to under <code>displayParent</code>.
	 *  <p>Method is called recursively.
	 */
	private void moveSubtree(DataNode child, DisplayNode displayParent)
	{
		if (child.displayNode != null)
		{
			child.displayNode.addUnderParent(displayParent);
			return;
		}

		else
		{
			Iterator iter = dataGraph.getEdges(child).iterator();
			while (iter.hasNext())
			{
				DataNode c = (DataNode) iter.next();
				moveSubtree(c, displayParent);
			}
		}
	}

	/** Policy for nodes with the same score value
	  * @see #repositionNodeInSortedList */
	private static final int INSERT_NODE_BEFORE = 0;

	/** Policy for nodes with the same score value
	 *  @see #repositionNodeInSortedList */
	private static final int INSERT_NODE_AFTER = 1;

	/** Moves <code>node</code> from its current position in <code>_sortedNodeList</code>
	 *  to a new location based on its current score.
	 *  <p>If <code>insertPolicy</code> is <code>INSERT_NODE_BEFORE</code>, the node is
	 *  inserted before all other nodes with the same score; if it is <code>INSERT_NODE_AFTER</code>
	 *  node is inserted after all other nodes with the same score.
	 *
	 *  @param node             node to be repositioned in <code>_sortedNodeList</code>
	 *  @param insertPolicy     one of <code>INSERT_NODE_BEFORE, INSERT_NODE_AFTER</code>
	 *  @return                 the new index of <code>node</code>
	 */
	private int repositionNodeInSortedList(DataNode node, int insertPolicy)
	{
		if (_sortedNodeList.remove(node))       // only reposition if the node was in the list
												// (the root is never in the list)
		{
			int i;
			for (i = 0; i < _sortedNodeList.size(); i++)
			{
				DataNode aNode = (DataNode) _sortedNodeList.get(i);
				if ( (aNode.score < node.score) ||
					(insertPolicy == INSERT_NODE_BEFORE && aNode.score == node.score))
						break;  // found the spot
			}

			_sortedNodeList.add(i, node);
			return i;
		}
		else
			return -1;
	}

	protected void finalize()
	{
		_debugOutput.close();
	}

	/** For debugging purposes--executes only if <code>debugLevel >= 3</code>.
	 *  Checks to ensure that only <code>numDisplayedNodes - 1</code> of
	 *  <code>_sortedNodeList</code> are displayed.
	 *  Otherwise, prints a message to <code>System.out</code>.
	 */
	private void doSanityCheck()
	{
		if (debugLevel < 3)
			return;

		for (int i = 0; i < _numDisplayedNodes; i++)
		{
			DataNode node = (DataNode)_sortedNodeList.get(i);
			if (node.displayNode == null)
				System.out.println("!!!! Node " + i + " that SHOULD BE displayed had displayNode == null ! : " + node);

		}

		int screwedUpIndex = -1;

		for (int i = _numDisplayedNodes; i < _sortedNodeList.size(); i++)
		{
			DataNode node = (DataNode)_sortedNodeList.get(i);
			if (node.displayNode != null)
			{
				System.out.println("!!!! Node " + i + " that should be NOT displayed had displayNode != null ! : " + node);
				screwedUpIndex = i;
			}
		}

		if (screwedUpIndex != -1)
		{
			_debugOutput.flush();
			printDisplayedNodes(new PrintWriter(System.out), screwedUpIndex + 1);
		}
	}

	/** Loads a property value that is a percentage, returns the value divided by 100. */
	private static float loadPercentProperty(String propertyName, int defaultValue)
	{
		return Integer.getInteger(propertyName, defaultValue).intValue() / 100.0f;
	}

	/** Println the debug string <code>message</code> if <code>messageLevel <= debugLevel</code> */
	private void debugPrintln(int messageLevel, String message)
	{
		if (messageLevel <= debugLevel)
			_debugOutput.println(message);
	}

	/** Prints the debug string <code>message</code> if <code>messageLevel <= debugLevel</code> */
	private void debugPrint(int messageLevel, String message)
	{
		if (messageLevel <= debugLevel)
			_debugOutput.print(message);
	}

	private void debugFlush(int messageLevel)
	{
		if (messageLevel <= debugLevel)
			_debugOutput.flush();
	}

	/** Creates a PrintWriter from a file with name specified by the environment variable
	 *  <code>filenameProperty</code>.
	 *  If file cannot be created, <code>System.out</code> is returned.
	 */
	private PrintWriter createDebugOutput(String filenameProperty, boolean autoFlush)
	{
		String filename = System.getProperty(filenameProperty);

		if (filename == null)
			return new PrintWriter(System.out, autoFlush);
		else
		{
			try { return new PrintWriter(new FileOutputStream(filename), autoFlush); }
			catch (IOException e)
			{
				e.printStackTrace();
				return new PrintWriter(System.out, autoFlush);
			}
		}
	}

	/** For debugging, executes only if <code>debugLevel < 5</code>. */
	private void printDisplayedNodes(PrintWriter out, int numNodesToPrint)
	{
		if (debugLevel < 5)
			return;

		out.println("-----");
		out.println("Displayed nodes: " + _numDisplayedNodes + "  - printing " + numNodesToPrint);

		for (int i = 0; i < numNodesToPrint; i++)
			out.println(_sortedNodeList.get(i));

		out.println();
		out.println();
		out.flush();
	}

	/** For debugging, executes only if <code>debugLevel < 5</code>. */
	private void printDisplayedNodes(PrintWriter out)
	{
		printDisplayedNodes(out, _numDisplayedNodes);
	}
}
